#include <QTRSensors.h>
#include <EEPROM.h>
#include <ServoTimer2.h>

//speed of OUR bluetooth adapter is 57600
//const uint8_t GLOBAL_DELAY = 2;
QTRSensors sensor;
ServoTimer2 tourbine;

const uint8_t SENSOR_BUF_SIZE = 8;
uint16_t SENSOR_BUF[SENSOR_BUF_SIZE];
const uint8_t ERROR_BUF_SIZE = 10;
int ERROR_BUF[ERROR_BUF_SIZE];

const uint8_t POS_BUF_SIZE = 10;
float POS_BUF[POS_BUF_SIZE];

const int B1_PIN = 11;
const int TOURBINE_PIN = 6;
const int L_MOTOR_EN_PIN = 3;
const int L_MOTOR_DIR_PIN = 4;
const int L_MOTOR_PWM_PIN = 10;
const int R_MOTOR_EN_PIN = 8;
const int R_MOTOR_DIR_PIN = 7;
const int R_MOTOR_PWM_PIN = 9;
int MAX_CROSSING_COUNT = 3;

bool B1_PRESSED = false;
bool PREPARED = false;
bool STARTED = false;
bool WAS_STARTED = false;

bool DEBUG_INFO = true;


unsigned long oldTime = 0;
unsigned long totalTime = 0;
unsigned long currHorLineTime = 0;
int crossingCount = 0;
boolean stopped = false;
boolean isOnCrossing = false;
boolean isOnBreak = false;
boolean wasOnBreak = false;
float oldRawPos = 0;
int breakCyclesCount = 0;
boolean stillOnCrossing = false;

int BREAK_CYCLES_MAX_COUNT = 50;

int TOURBINE_ON_SPEED = 800;  //      900
int TOURBINE_OFF_SPEED = 800;

//int BASE_SPEED = 105;
//int MAX_SPEED = 170;

int BASE_SPEED = 100;
int MAX_SPEED = 200;


boolean NO_MOVE = false;

int errBufPos = -1;
int posBufPos = -1;


//float kP = 0.25; // коэффициент пропорциональной обратной связи
//float kD = 0.57; // коэффициент дифференциальной обратной связи

float kP = 0.0784; // коэффициент пропорциональной обратной связи   0.078      0.080    0.0784
float kD = 0.01986; // коэффициент дифференциальной обратной связи   0.017      0.020    0.0198
float kI = 0.00831; // коэффициент интегральной обратной связи      0.51       0.0082   0.00829

//float kP = 0.070; // коэффициент пропорциональной обратной связи   0.078
//float kD = 0.015; // коэффициент дифференциальной обратной связи   0.017
//float kI = 0.01; // коэффициент интегральной обратной связи      0.51


void setup()
{
  Serial.begin(57600);
  setupHardware();
  motorsForward(0);
  doCalibration();
  Serial.println("PRESS B1 or ENTER TO PREPARE, THEN PRESS B1 or ENTER TO START");
  Serial.println("PRESS B1 or ENTER AGAIN TO STOP");
}

void loop() {
  if (STARTED) {
    WAS_STARTED = true;
    float error = bot_position(); // Это ошибка положения
    if (isOnCrossing && !stillOnCrossing) {
      crossingCount++;
      unsigned long currTime = millis() - oldTime;
      totalTime += currTime;
      Serial.println("Round number: " + String(crossingCount) + ", Round Time: " + String(currTime));
      if (crossingCount >= MAX_CROSSING_COUNT) {
        stopped = true;
        motorsForward(0, 0);
      }
      stillOnCrossing = true;
      oldTime = millis();
    }
    if (stillOnCrossing) {
      if (millis() - oldTime > 1000) {
        stillOnCrossing = false;
      }
    }
    float K = kP * error;
    errBufPos = (errBufPos + 1) % ERROR_BUF_SIZE;
    ERROR_BUF[errBufPos] = error;
    float D = (error - ERROR_BUF[(errBufPos + ERROR_BUF_SIZE + 1) % ERROR_BUF_SIZE]) * kD;

    int errorSum = 0;
    for (int i = 0; i < ERROR_BUF_SIZE; i++) errorSum += ERROR_BUF[i];
    float I = errorSum / ERROR_BUF_SIZE * kI;

    float correction =  K + I + D;

    int leftSpeed = BASE_SPEED * (1 - correction);
    int rightSpeed = BASE_SPEED * (1 + correction);
    int ls = leftSpeed;
    int rs = rightSpeed;

    //    if (ls > MAX_SPEED) {
    //      rs = rs - (ls - MAX_SPEED);
    //      ls = MAX_SPEED;
    //    }
    //
    //    if (rs > MAX_SPEED) {
    //      ls = ls - (rs - MAX_SPEED);
    //      rs = MAX_SPEED;
    //    }

    leftSpeed = constrain(ls, 0, MAX_SPEED);
    rightSpeed = constrain(rs, 0, MAX_SPEED);
    if (NO_MOVE) {
      motorsForward(0, 0);
      tourbine.write(TOURBINE_OFF_SPEED);
    } else {
      tourbine.write(TOURBINE_ON_SPEED);
      motorsForward(leftSpeed, rightSpeed);
    }
    if (DEBUG_INFO) {
      String s = " POS:" + String(error) + " CORR:" + String(correction) + " L:" + String(leftSpeed) + " R:" + String(rightSpeed);
      //Serial.println(s);
    }
    delay(0);
  } else {
    if (WAS_STARTED) {
      tourbine.write(TOURBINE_OFF_SPEED);
      WAS_STARTED = false;
      PREPARED = false;
    }
    motorsForward(0, 0);
  }


  if (digitalRead(B1_PIN) == HIGH) {
    Serial.println("B1_PRESSED");
    B1_PRESSED = true;
    delay(400);
  }


  if ((Serial.available() > 0 && Serial.read() == 10) || B1_PRESSED) {
    B1_PRESSED = false;
    if (!PREPARED && !STARTED) {
      PREPARED = true;
      Serial.println("BOT is PREPARED;");
    }
    else if (PREPARED && !STARTED)
    {
      STARTED = true;
      Serial.println("BOT is STARTED");
    }
    else {
      STARTED = false;
      PREPARED = false;
      Serial.println("ALL STOPPED");
    }
  }
  if (!NO_MOVE && PREPARED) {
    tourbine.write(TOURBINE_ON_SPEED);
    delay(10);
  }
}


float bot_position() {
  int sensorValue = (sensor.readLineBlack(SENSOR_BUF) - 3500);
  float rawPos = sensorValue / 1000.00;

  //определяем что мы на пересечении стартовой линии
  int sensorValuesSum = 0;
  for (int i = 0; i < SENSOR_BUF_SIZE; i++) {
    sensorValuesSum += SENSOR_BUF[i];
  }
  isOnCrossing = sensorValuesSum > 4000;

  //заполняем буфер предыдущими значениями датчика
  posBufPos = (posBufPos + 1) % POS_BUF_SIZE;
  POS_BUF[posBufPos] = rawPos;

  //получаем предыдущее значение датчика на POS_BUF_SIZE шагов назад
  float prevRawPos =  POS_BUF[(posBufPos + POS_BUF_SIZE + 1) % POS_BUF_SIZE];
  float k = abs((rawPos - prevRawPos) / POS_BUF_SIZE) * 10;

  //if (k >= 7) rawPos = prevRawPos;

  //мы на перерыве если:
  //под всем датчиками нет линии (sensorValuesSum < 500)
  //датчик показывает крайнее значение 3.5
  //предыдущее значение было в пределах от -2 до 2
  isOnBreak = (sensorValuesSum < 500 && abs(rawPos) >= 3.5 && prevRawPos > -2 &&  prevRawPos < 2);

  //если определили что наехали на перерыв, то запоминаем что мы на перерыве и последнее значение датчика
  //  if (isOnBreak && !wasOnBreak) {
  //    wasOnBreak = true;
  //    oldPrevRawPos = prevRawPos;
  //  }

  //если мы на перерые то сохраняем состояние на перерыве заданное кол-ва циклов
  //  if (wasOnBreak) {
  //    breakCyclesCount++;
  //    if (breakCyclesCount > BREAK_CYCLES_MAX_COUNT) {
  //      isOnBreak = false;
  //      wasOnBreak = false;
  //      breakCyclesCount = 0;
  //    }
  //  }

  int bbb = wasOnBreak || isOnBreak ? 5 : 2;
  String s1 = String(abs(rawPos));
  s1.replace(".", ",");
  String s2 = String(prevRawPos);
  s2.replace(".", ",");
  String s3 = String(bbb);

  //Serial.println(s1 + " " + s2 + " " + s3);

  //получаем усреднённое с учётом предыдущих значений положение
  //  int idx = posBufPos;
  //  float rawPosSum = 0;
  //  for (int i = 0; i < POS_BUF_SIZE; i++) {
  //    rawPosSum += POS_BUF[idx];
  //    idx--;
  //    if (idx < 0) idx = POS_BUF_SIZE - 1;
  //  }
  //
  //  float avgRawPos = isOnBreak || wasOnBreak ? oldPrevRawPos : rawPosSum / POS_BUF_SIZE;

  //float pos3 = pos2 + (pos2 * pos2 ) / 10 * sgn(pos2);
  //float poweredPos = avgRawPos + (avgRawPos * avgRawPos ) / 3 * sgn(avgRawPos);

  //  float storedRawPos = isOnBreak ? oldPrevRawPos : rawPos;
  //  float poweredPos = storedRawPos + (storedRawPos * storedRawPos ) / 3 * sgn(storedRawPos);
  float poweredPos = rawPos + (rawPos * rawPos ) / 6 * sgn(rawPos);
  s1 = String(rawPos);
  s1.replace(".", ",");
  s2 = String(poweredPos);
  s2.replace(".", ",");
  s3 = String(k);
  s3.replace(".", ",");
  Serial.println(String(s1) + " " + String(s2) + " " + String(s3));
  oldRawPos = rawPos;
  return poweredPos;
}

static inline int8_t sgn(int val) {
  if (val < 0) return -1;
  if (val == 0) return 0;
  return 1;
}


void doCalibration() {
  Serial.println("B1 or ENTER FOR RE-CALIBRATION");
  for (int i = 0; i < 400; i++) {
    if (i % 10 == 0)
      digitalWrite(LED_BUILTIN, HIGH);
    if (digitalRead(B1_PIN) == HIGH) {
      B1_PRESSED = true;
      delay(300);
    }
    if ((Serial.available() > 0 && Serial.read() == 10) || B1_PRESSED) {
      B1_PRESSED = false;
      digitalWrite(LED_BUILTIN, HIGH);
      Serial.print("CALIBRATION: ");
      for (uint16_t i = 0; i < 400; i++) {
        sensor.calibrate();
        for (int i = 0; i < SENSOR_BUF_SIZE; i++) {
          Serial.print("[" + String(sensor.calibrationOn.minimum[i]) + ":" + String(sensor.calibrationOn.maximum[i]) + "] ");
        }
        Serial.println();
      }
      digitalWrite(LED_BUILTIN, LOW);
      saveCalibration();
      break;
    }
    if (i % 20 == 0)
      digitalWrite(LED_BUILTIN, LOW);
    delay(10);
  }
  delay(100);
  loadCalibration();
  delay(100);
}

void saveCalibration() {
  Serial.print("SAVING CALIBRATION: ");
  for (int i = 0; i < SENSOR_BUF_SIZE; i++) {
    Serial.print("[" + String(sensor.calibrationOn.minimum[i]) + ":" + String(sensor.calibrationOn.maximum[i]) + "]");
  }
  Serial.print("...");
  for (int i = 0; i < SENSOR_BUF_SIZE; i++) {
    EEPROM.put(i * 2, sensor.calibrationOn.minimum[i]);
    EEPROM.put(i * 2 + SENSOR_BUF_SIZE * 2, sensor.calibrationOn.maximum[i]);
  }
  Serial.println(" DONE");
  delay(1000);
}

void loadCalibration() {
  Serial.print("LOADING CALIBRATION: ");
  sensor.calibrate();  //init calibration array first
  for (int i = 0; i < SENSOR_BUF_SIZE; i++) {
    EEPROM.get(i * 2, sensor.calibrationOn.minimum[i]);
    EEPROM.get(i * 2 + SENSOR_BUF_SIZE * 2, sensor.calibrationOn.maximum[i]);
  }
  for (int i = 0; i < SENSOR_BUF_SIZE; i++) {
    Serial.print("[" + String(sensor.calibrationOn.minimum[i]) + ":" + String(sensor.calibrationOn.maximum[i]) + "]");
  }
  Serial.println(" DONE");
  delay(1000);
}


void setupHardware() {
  Serial.print("SETUP HARDWARE...");
  pinMode(B1_PIN, INPUT);
  //motors
  pinMode(L_MOTOR_DIR_PIN, OUTPUT);
  pinMode(R_MOTOR_DIR_PIN, OUTPUT);
  pinMode(L_MOTOR_EN_PIN, OUTPUT);
  pinMode(R_MOTOR_EN_PIN, OUTPUT);
  TCCR1A = 0b00000001;  // 8bit
  TCCR1B = 0b00000001;
  motorsForward(0);

  //tourbine
  tourbine.attach(TOURBINE_PIN);
  tourbine.write(800);

  //sensors
  sensor.setTypeAnalog();
  sensor.setSamplesPerSensor(1);
  sensor.setSensorPins((const uint8_t[]) {
    A0, A1, A2, A3, A4, A5, A6, A7
  }, SENSOR_BUF_SIZE);

  for (int i = 0; i < SENSOR_BUF_SIZE; i++)
    SENSOR_BUF[i] = 0;
  for (int i = 0; i < ERROR_BUF_SIZE; i++)
    ERROR_BUF[i] = 0;

  Serial.println("DONE");
}

void motorsForward(int leftSpeed, int rightSpeed) {
  //use electronoic breaks
  if (leftSpeed == 0) digitalWrite(L_MOTOR_EN_PIN, HIGH); else digitalWrite(L_MOTOR_EN_PIN, LOW);
  if (rightSpeed == 0) digitalWrite(R_MOTOR_EN_PIN, HIGH); else digitalWrite(R_MOTOR_EN_PIN, LOW);
  digitalWrite(L_MOTOR_DIR_PIN, HIGH);
  digitalWrite(R_MOTOR_DIR_PIN, LOW);
  analogWrite(L_MOTOR_PWM_PIN, leftSpeed);
  analogWrite(R_MOTOR_PWM_PIN, rightSpeed);
}

void motorsForward(int motorsSpeed) {
  motorsForward(motorsSpeed, motorsSpeed);
}

void motorsBackward(int leftSpeed, int rightSpeed) {
  digitalWrite(L_MOTOR_DIR_PIN, LOW);
  digitalWrite(R_MOTOR_DIR_PIN, HIGH);
  analogWrite(L_MOTOR_PWM_PIN, leftSpeed);
  analogWrite(R_MOTOR_PWM_PIN, rightSpeed);
}

void motorsBackward(int motorsSpeed) {
  motorsBackward(motorsSpeed, motorsSpeed);
}
